/**
 * 
 */
package org.gcube.accounting.persistence;

import java.util.Calendar;
import java.util.concurrent.TimeUnit;

import org.gcube.accounting.datamodel.SingleUsageRecord;
import org.gcube.accounting.datamodel.basetypes.TestUsageRecord;
import org.gcube.accounting.exception.InvalidValueException;
import org.gcube.accounting.testutility.StressTestUtility;
import org.gcube.accounting.testutility.TestOperation;
import org.gcube.common.scope.api.ScopeProvider;
import org.junit.Assert;
import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author Luca Frosini (ISTI - CNR) http://www.lucafrosini.com/
 *
 */
public class AccountingPersistenceBackendTest {
	
	private static Logger logger = LoggerFactory.getLogger(AccountingPersistenceBackendTest.class);
	
	public static final String[] SCOPES = new String[]{"/gcube", "/gcube/devNext"};
	public static final String GCUBE_SCOPE = SCOPES[0];
	public static final String GCUBE_DEVNEXT_SCOPE = SCOPES[1];
	
	public static final long timeout = 5000;
	public static final TimeUnit timeUnit = TimeUnit.MILLISECONDS;
	
	public static AccountingPersistenceBackend getPersistence(){
		ScopeProvider.instance.set(GCUBE_DEVNEXT_SCOPE);
		AccountingPersistenceBackendFactory.setFallbackLocation(null);
		return AccountingPersistenceBackendFactory.getPersistenceBackend();
	}
	
	@Test
	public void singleTestNoScope() throws Exception {
		AccountingPersistenceBackendFactory.setFallbackLocation(null);
		final AccountingPersistenceBackend persistence = AccountingPersistenceBackendFactory.getPersistenceBackend();
		Assert.assertTrue(persistence instanceof FallbackPersistenceBackend);
		StressTestUtility.stressTest(new TestOperation() {
			@Override
			public void operate(int i) {
				SingleUsageRecord usageRecord;
				try {
					usageRecord = TestUsageRecord.createTestServiceUsageRecordExplicitScope();
					persistence.validateAccountAggregate(usageRecord, true, false);
				} catch (InvalidValueException e) {
					throw new RuntimeException(e);
				}
			}
		}, 1);
		
		persistence.flush(timeout, timeUnit);
	}
	
	@Test
	public void singleTest() throws Exception {
		final AccountingPersistenceBackend persistence = getPersistence();
		StressTestUtility.stressTest(new TestOperation() {
			@Override
			public void operate(int i) {
				SingleUsageRecord usageRecord = TestUsageRecord.createTestServiceUsageRecordAutomaticScope();
				persistence.validateAccountAggregate(usageRecord, true, false);
			}
		}, 1);
		
		persistence.flush(timeout, timeUnit);
	}
	
	@Test
	public void stressTestNoAggregation() throws Exception {
		final AccountingPersistenceBackend persistence = getPersistence();
		StressTestUtility.stressTest(new TestOperation() {
			@Override
			public void operate(int i) {
				SingleUsageRecord usageRecord = TestUsageRecord.createTestServiceUsageRecordAutomaticScope();
				persistence.validateAccountAggregate(usageRecord, true, false);
			}
		});
	}
	
	@Test
	public void stressTestWithAggregation() throws Exception {
		final AccountingPersistenceBackend persistence = getPersistence();
		
		StressTestUtility.stressTest(new TestOperation() {
			@Override
			public void operate(int i) throws Exception {
				SingleUsageRecord usageRecord = TestUsageRecord.createTestServiceUsageRecordAutomaticScope();
				persistence.account(usageRecord);
			}
		});
		persistence.flush(timeout, timeUnit);
	}
	
	
	@Test
	public void testScopeRecheck() throws Exception {
		ScopeProvider.instance.set("/fakeScope");
		logger.debug("Going to check First Time");
		AccountingPersistenceBackend first = AccountingPersistenceBackendFactory.getPersistenceBackend();
		logger.debug("First {} : {}", AccountingPersistenceBackend.class.getSimpleName(), first);
		Long firstCheck = AccountingPersistenceBackendFactory.getFallbackLastCheck(ScopeProvider.instance.get());
		logger.debug("First Check Time {}", firstCheck);
		
		long startTime = Calendar.getInstance().getTimeInMillis();
		long endTime = startTime;
		
		while(endTime <=  (startTime + (AccountingPersistenceBackendFactory.FALLBACK_RETRY_TIME + 100))){
			endTime = Calendar.getInstance().getTimeInMillis();
		}

		logger.debug("Going to check Second Time");
		AccountingPersistenceBackend second = AccountingPersistenceBackendFactory.getPersistenceBackend();
		logger.debug("Second {} : {}", AccountingPersistenceBackend.class.getSimpleName(), second);
		Long secondCheck = AccountingPersistenceBackendFactory.getFallbackLastCheck(ScopeProvider.instance.get());
		logger.debug("Second Check Time {}", secondCheck);
		
		Assert.assertNotEquals(firstCheck, secondCheck);
		Assert.assertEquals(first, second);
	}
	
}
